//
//  Model.m
//  Distance
//
//  Created by NYU User on 3/27/11.
//  Copyright 2011 __MyCompanyName__. All rights reserved.
//

#import "Model.h"
#import <CoreLocation/CoreLocation.h>

//Return the distance in miles between two locations.
#define METERS_PER_MILE 1609.344
const int n = 4;	//number of arguments of sqlite distance function

static void distance(sqlite3_context *context, int argc, sqlite3_value **argv) {
	if (argc != n) {
		static char buffer[128];
		sprintf(buffer, "distance requires %d arguments", n);;
		sqlite3_result_error(context, buffer, strlen(buffer));
		return;
	}
	
	double a[n];	//Fill this with the n arguments.
	
	for (int i = 0; i < argc; ++i) {
		int t = sqlite3_value_type(argv[i]);
		if (t != SQLITE_FLOAT) {
			NSLog(@"sqlite3_value_type(argv[%d]) == %d", i, t);
			char *error = "distance requires real arguments";
			sqlite3_result_error(context, error, strlen(error));
			return;
		}
		a[i] = sqlite3_value_double(argv[i]);
	}
	
	CLLocation *loc1 = [[CLLocation alloc] initWithLatitude: a[0] longitude: a[1]];
	CLLocation *loc2 = [[CLLocation alloc] initWithLatitude: a[2] longitude: a[3]];
	
	double meters = [loc1 distanceFromLocation: loc2];
	[loc2 release];
	[loc1 release];
	double miles = meters / METERS_PER_MILE;
	sqlite3_result_double(context, miles);
}

@implementation Model

- (id) init {
	if ((self = [super init]) != nil) {
		//Open the databse.
		NSBundle *bundle = [NSBundle mainBundle];
		NSString *fileName = [bundle pathForResource: @"zipcodes" ofType: @"db"];
		if (sqlite3_open([fileName UTF8String], &base) != SQLITE_OK) {
			NSLog(@"sqlite3_open: %s", sqlite3_errmsg(base));
			[self release];
			return nil;
		}
		
		if (sqlite3_create_function(
			base,
			"distance",		//name of sql function
			n,				//number of arguments
			SQLITE_UTF8,	//text encoding
			NULL,
			distance,		//pointer to C function
			NULL,
			NULL
		) != SQLITE_OK) {
			NSLog(@"sqlite3_create_function failed.");
			[self release];
			return nil;
		}
	}
	return self;
}

- (NSMutableString *) getZipCodes: (NSString *) zipCode {

	//the return value
	NSMutableString *zipCodes = [NSMutableString stringWithCapacity: 0];

	//Get the latitude and longitude of the target zipcode.
	
	char *sql = sqlite3_mprintf(
		"select latitude, longitude from zipcodes where zipcode == %s;",
		[zipCode UTF8String]
	);
	
	sqlite3_stmt *statement;
	
	if (sqlite3_prepare_v2(base, sql, strlen(sql), &statement, NULL)
		!= SQLITE_OK) {
		NSLog(@"%s", sqlite3_errmsg(base));
		return zipCodes;
	}
	
	sqlite3_free(sql);
	
	if (sqlite3_step(statement) != SQLITE_ROW) {
		NSLog(@"Could not find zipcode %s.", zipCode);
		return zipCodes;
	}
	
	NSAssert(sqlite3_column_type(statement, 0) == SQLITE_FLOAT,
		@"column 0 (latitude) must be SQLITE_FLOAT.");
	NSAssert(sqlite3_column_type(statement, 1) == SQLITE_FLOAT,
		@"column 1 (longitude) must be SQLITE_FLOAT.");
	
	double  latitude = sqlite3_column_double(statement, 0);
	double longitude = sqlite3_column_double(statement, 1);
	
	if (sqlite3_finalize(statement) != SQLITE_OK) {
		NSLog(@"sqlite3_finalized failed.");
	}

	//Get all the zipcodes that are close to the target zipcode.

	sql = sqlite3_mprintf(
		"select *, distance(latitude, longitude, %.15g, %.15g) as d "
		"from zipcodes "
		"where d < 5 "
		"order by d;",
		latitude, longitude
	);
	
	if (sqlite3_prepare_v2(base, sql, strlen(sql), &statement, NULL)
		!= SQLITE_OK) {
		NSLog(@"%s", sqlite3_errmsg(base));
		return zipCodes;
	}
	
	sqlite3_free(sql);
	
	for (int rank = 0; sqlite3_step(statement) == SQLITE_ROW; ++rank) {
		NSAssert(sqlite3_column_type(statement, 0) == SQLITE_INTEGER,
			@"column 0 must be SQLITE_INTEGER.");
		NSAssert(sqlite3_column_type(statement, 1) == SQLITE3_TEXT,
			@"column 1 must be SQLITE3_TEXT.");
		NSAssert(sqlite3_column_type(statement, 2) == SQLITE3_TEXT,
			@"column 2 must be SQLITE3_TEXT.");
		NSAssert(sqlite3_column_type(statement, 3) == SQLITE_FLOAT,
			@"column 3 must be SQLITE_FLOAT.");
		NSAssert(sqlite3_column_type(statement, 4) == SQLITE_FLOAT,
			@"column 4 must be SQLITE_FLOAT.");
		NSAssert(sqlite3_column_type(statement, 5) == SQLITE3_TEXT,
			@"column 5 must be SQLITE3_TEXT.");
		NSAssert(sqlite3_column_type(statement, 6) == SQLITE_FLOAT,
			@"column 6 must be SQLITE_FLOAT.");
		
		[zipCodes appendFormat:
			@"%05d %s, %s\n"
			"rank: %d\n"
			"county: %s\n"
			"lat/long: %g/%g\n"
			"miles: %g\n\n",
		 
			sqlite3_column_int(statement, 0),		//zipcode
			sqlite3_column_text(statement, 1),		//city
			sqlite3_column_text(statement, 2),		//state
			rank,
			sqlite3_column_text(statement, 5),		//county
			sqlite3_column_double(statement, 3),	//latitude
			sqlite3_column_double(statement, 4),	//longitude
			sqlite3_column_double(statement, 6)		//distance
		];
	}
	
	if (sqlite3_finalize(statement) != SQLITE_OK) {
		NSLog(@"sqlite3_finalized failed.");
	}

	return zipCodes;
}

- (void) dealloc {
	if (sqlite3_close(base) != SQLITE_OK) {
		NSLog(@"sqlite3_close failed.");
	}
	
	[super dealloc];
}

@end
